unit NWHPrint;

{
  HostPrinter for Netware component (C)2001-2007 by Paul TOTH <tothpaul@free.fr>
  http://tothpaul.free.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  Novell;

type
  TNWHostPrinter = class(TComponent)
  private
  // socket
   fActive  :boolean;
   fHostName:string;
   fState   :integer;
   fData    :string;
   fNegoce  :string;
  // windows
   fHandle  :THandle;
   fPrinter :string;
   fSocket  :integer;
  { novell }
   fContext :integer;
   fConnect :integer;
   fQueue   :integer;
   fJob     :TNWQueueJob;
  { ------ }
   hPrinter :integer;
   ibmcode  :byte;
   hpcl     :string;
  // IBM
   fDevName :string;
   fMSGQ    :string;
   fModel   :string;
   fWSCST   :string;
   fPaper1  :string;
   fPaper2  :string;
   fEnvelope:string;
   fFont    :string;
   fFormFeed:string;
   fRemoteName:string;
  // general
   fError   :integer;
   fErrorMsg:string;
   fStatus  :string;
   EOnConnect:TNotifyEvent;
   EOnChange :TNotifyEvent;
   EOnError  :TNotifyEvent;
   EOnClose  :TNotifyEvent;
   EOnPrint  :TNotifyEvent;
   EOnPrintError:TNotifyEvent;
  // socket
   procedure SetActive(Value:boolean);
   procedure WndProc(Var Msg:TMessage);
   procedure ReadData(Socket:integer);
   procedure Disconnected;
   function  Execute:string;
   procedure StartupInfo;
   function  PrintRecord:string;
  protected
    { Dclarations protges }
   procedure SetError(Lvl:integer; Msg:string);
   function GetError:integer;
   procedure SetStatus(Value:string);
   function GetPrinting:boolean;
  // Novell
   function OpenPrinter:boolean;
   function WritePrinter(s:string):boolean;
   function AbortDoc:boolean;
   function EndDoc:boolean;
  public
    { Dclarations publiques }
   Constructor Create(AOwner:TComponent); override;
   Destructor Destroy; override;
   procedure Loaded; override;
   procedure connect;
   procedure Disconnect;
  // Setup
   Function Setup:boolean; // Return OK/Cancel
  // general
   property Error   :integer read GetError;
   property ErrorMsg:string  read fErrorMsg;
   property Status  :string  read fStatus write SetStatus;
   property Printing :boolean read GetPrinting;
  published
  // socket
   property Active  :boolean read fActive  write SetActive;
   property HostName:string read fHostName write fHostName;
  // Novell
   property Printer:string read fPrinter write fPrinter;
  // IBM
   property DevName :string read fDevName  write fDevName;
   property MSQG    :string read fMSGQ     write fMSGQ;
   property Model   :string read fModel    write fModel;
   property WSCST   :string read fWSCST    write fWSCST;
   property Paper1  :string read fPaper1   write fPaper1;
   property Paper2  :string read fPaper2   write fPaper2;
   property Envelope:string read fEnvelope write fEnvelope;
   property Font    :string read fFont     write fFont;
   property FormFeed:string read fFormFeed write fFormFeed;
  // General
   property OnConnect:TNotifyEvent read EOnConnect write EOnConnect;
   property OnChange :TNotifyEvent read EOnChange  write EOnChange;
   property OnError  :TNotifyEvent read EOnError   write EOnError;
   property OnClose  :TNotifyEvent read EOnClose   write EOnClose;
   property OnPrint  :TNotifyEvent read EOnPrint   write EOnPrint;
   property OnPrintError:TNotifyEvent read EOnPrintError write EOnPrintError;
  end;

//procedure Register;

function GetName(Name:string):string; // return Name in a string [LIB/]NAME
function GetLib(Name:string):string; // return LIB (or *LIBL) in a string [LIB/]NAME

implementation

Uses
 Winsock,EBCDIC,NWHSETUP;//dsgnintf;

function GetName(Name:string):string; // return Name in a string [LIB/]NAME
var
 p:integer;
begin
 p:=pos('/',Name);
 if p=0 then Result:=Name else Result:=Copy(Name,p+1,Length(Name));
end;

function GetLib(Name:string):string; // return LIB (or *LIBL) in a string [LIB/]NAME
var
 p:integer;
begin
 p:=pos('/',Name);
 if p=0 then Result:='*LIBL' else Result:=Copy(Name,1,p-1);
end;


(*///////////////// Telnet Const ///////////////////////*)
const
 TERM  = 'IBM-3812-1';

 tnEND      =#239; // End of 5250 DataStream
 tnSE       =#240;
 tnNOP      =#241;
 tnDATA     =#242;
 tnBRK      =#243;
 tnINT      =#244;
 tnABORT    =#245;
 tnAYT      =#246; // Are you there
 tnERASECHAR=#247;
 tnERASELINE=#248;
 tnGOAHEAD  =#249;
 tnSB       =#250;
 tnWILL     =#251;
 tnWONT     =#252;
 tnDO       =#253;
 tnDONT     =#254;
 tnIAC      =#255;

 tnBIN        =#0;
 tnECHO       =#1;
 tnSUPPRESS_GA=#3;
 tnTIMING     =#6;

 tnTERMTYPE   =#24; // Terminal type
 tnEOR        =#25;
 tnNEWENV     =#39; // Session name

 ttIS      =#0;
 ttSEND    =#1;
 ttINFO    =#2;

 tnVAR     =#0;
 tnVAL     =#1;
(*//////////////////////////////////////////////////////*)

Const
 Papers:array[0..12] of PChar=(
  '*NONE', '*MFRTYPMDL', '*LETTER', '*LEGAL', '*EXECUTIVE',
  '*A4', '*A5', '*B5', '*CONT80', '*CONT132', '*A3', '*B4', '*LEDGER'
 );
 Envelops:array[0..7] of PChar=(
  '*NONE', '*MFRTYPMDL', '*B5', '*MONARCH', '*NUMBER9', '*NUMBER10', '*C5', '*DL'
 );

(*///////////////// Property Editors ///////////////////*)
(*
Type
 TValueListProperty=class(TStringProperty)
  public
   Function GetAttributes:TPropertyAttributes; override;
 end;

Function TValueListProperty.GetAttributes;
 begin
  Result:=[paValueList];
 end;

Type
 TModelProperty=class(TValueListProperty)
  public
   Procedure GetValues(Proc:TGetStrProc); override;
 end;

Procedure TModelProperty.GetValues(Proc:TGetStrProc);
 begin
 {
      { *SAME    }     Proc('*IBM39303');
  Proc('*IBM2380');    Proc('*IBM4019');
  Proc('*IBM2381');    Proc('*IBM4019HP');
  Proc('*IBM2390');    Proc('*IBM4029');
  Proc('*IBM2391');    Proc('*IBM4029HP');
  Proc('*IBM3112');    Proc('*IBM4037');
  Proc('*IBM3116');    Proc('*IBM4039HP');
  Proc('*IBM3130');    Proc('*IBM4070');
  Proc('*IBM3812');    Proc('*IBM4070EP');
  Proc('*IBM3816');    Proc('*IBM4072');
  Proc('*IBM3912HP');  Proc('*IBM4076');
  Proc('*IBM3916HP');  Proc('*IBM42011');
  Proc('*IBM39302');   Proc('*IBM42012');
 //---
  Proc('*IBM42013');   Proc('*IBM4244MAN');
  Proc('*IBM42021');   Proc('*IBM4244ASF');
  Proc('*IBM42022');   Proc('*IBM4244DUAL');
  Proc('*IBM42023');   Proc('*IBM4247MAN');
  Proc('*IBM42071');   Proc('*IBM4247ASF');
  Proc('*IBM42072');   Proc('*IBM4247DUAL');
  Proc('*IBM42081');   Proc('*IBM4312');
  Proc('*IBM42082');   Proc('*IBM4317');
  Proc('*IBM4212');    Proc('*IBM4320');
  Proc('*IBM4216');    Proc('*IBM4324');
  Proc('*IBM4226');    Proc('*IBM4332');
  Proc('*IBM4230');    Proc('*IBM47121');
  Proc('*IBM4232');    Proc('*IBM47122');
 //---
  Proc('*IBM47221');   Proc('*IBM6404');
  Proc('*IBM47222');   Proc('*IBM6404EP');
  Proc('*IBM4770');    Proc('*IBM6408');
  Proc('*IBM5152');    Proc('*IBM6408EP');
  Proc('*IBM5201');    Proc('*IBM6412');
  Proc('*IBM5202');    Proc('*IBM6412EP');
  Proc('*IBM5204');    Proc('*CPQPM15');
  Proc('*IBM5216');    Proc('*CPQPM20');
  Proc('*IBM5575');    Proc('*HPII');
  Proc('*IBMPAGES');   Proc('*HPIID');
  Proc('*IBMPAGES300');Proc('*HPIIP');
  Proc('*IBM6400');    Proc('*HPIII');
  Proc('*IBM6400EP');  Proc('*HPIIID');
 //---
  Proc('*HPIIIP');     Proc('*HP310');
  Proc('*HPIIISI');    Proc('*HP320');
  Proc('*HP4');        Proc('*HP500');
  Proc('*HP5');        Proc('*HP520');
  Proc('*HP5SI');      Proc('*HP540');
  Proc('*HP6');        Proc('*HP550C');
  Proc('*HP4000');     Proc('*HP560C');
  Proc('*HP5000');     Proc('*HPPAINT');
  Proc('*HP8000');     Proc('*LEXOPTRA');
  Proc('*HPDBCS');     Proc('*LEXOPTRAC');
  Proc('*HPCOLORLJ');  Proc('*LEXOPTRAN');
  Proc('*HP1200C');    Proc('*LEXOPTRAS');
  Proc('*HP1600C');    Proc('*LEXOPTRASC');
 //---
  ...
 }

  Proc('*HPII');    Proc('*HPIID');  Proc('*HPIIP');   Proc('*HPIII');
  Proc('*HPIIID');  Proc('*HPIIIP'); Proc('*HPIIISI'); Proc('*HP4');
  Proc('*HP5');     Proc('*HP5SI');  Proc('*HP6');     Proc('*HP310');
  Proc('*HP320');   Proc('*HP500');  Proc('*HP520');   Proc('*HP540');
  Proc('*HP550C');  Proc('*HP560C'); Proc('*HPPAINT'); Proc('*EPFX850');
  Proc('*EPFX870'); Proc('*IBM2391');
  Proc('*WSCST');

 end;

Type
 TPaperSourceProperty=class(TValueListProperty)
  public
   Procedure GetValues(Proc:TGetStrProc); override;
 end;

Procedure TPaperSourceProperty.GetValues(Proc:TGetStrProc);
 var
  i:integer;
 begin
  for i:=Low(Papers) to High(Papers) do Proc(Papers[i]);
 end;

Type
 TEnvelopeProperty=class(TValueListProperty)
  public
   procedure GetValues(Proc:TGetStrProc); override;
 end;

Procedure TEnvelopeProperty.GetValues(Proc:TGetStrProc);
 var
  i:integer;
 begin
  for i:=Low(Envelops) to High(Envelops) do Proc(Envelops[i]);
 end;

Type
 TNWHostPrinterEditor=class(TComponentEditor)
  function GetVerbCount:integer; override;
  function GetVerb(Index:integer):string; override;
  procedure ExecuteVerb(Index:integer); override;
 end;

function TNWHostPrinterEditor.GetVerbCount:integer;
 begin
  Result:=1;
 end;

function TNWHostPrinterEditor.GetVerb(Index:integer):string;
 begin
  Result:='Setup';
 end;

procedure TNWHostPrinterEditor.ExecuteVerb(Index:integer);
 begin
  TNWHostPrinter(component).Setup;
 end;
***)

(*//////////////////////////////////////////////////////*)
(*
procedure Register;
begin
  RegisterComponents('LookInside.IBM', [TNWHostPrinter]);
//  RegisterPropertyEditor(TypeInfo(string),TNWHostPrinter,'Printer',TPrinterProperty);
  RegisterPropertyEditor(TypeInfo(string),TNWHostPrinter,'Model',TModelProperty);
  RegisterPropertyEditor(TypeInfo(string),TNWHostPrinter,'Paper1',TPaperSourceProperty);
  RegisterPropertyEditor(TypeInfo(string),TNWHostPrinter,'Paper2',TPaperSourceProperty);
  RegisterPropertyEditor(TypeInfo(string),TNWHostPrinter,'Envelope',TEnvelopeProperty);
  RegisterComponentEditor(TNWHostPrinter,TNWHostPrinterEditor);
end;
*)
Constructor TNWHostPrinter.Create(AOwner:TComponent);
 begin
  inherited;
  fMSGQ:='*LIBL/QSYSOPR';
  fModel:='*HPIII';
  fPaper1:='*A4';
  fPaper2:='*NONE';
  fEnvelope:='*NONE';
  fFont:='12';
  fFormFeed:='C';
 end;

Destructor TNWHostPrinter.Destroy;
 begin
  Disconnect;
  inherited;
 end;

Procedure TNWHostPrinter.Loaded;
 begin
  inherited;
  if fActive then begin
   fActive:=false;
   Connect;
  end;
 end;

Procedure TNWHostPrinter.SetActive(Value:boolean);
 begin
  if (csLoading in ComponentState) then
   fActive:=Value
  else
   if Value then Connect else Disconnect;
 end;

Procedure TNWHostPrinter.SetError(Lvl:integer; Msg:string);
 begin
  fError:=lvl;
  fErrorMsg:=Msg;
  if Assigned(EOnError) then
   EOnError(Self)
  else
   raise Exception.create(Name+':'+ClassName+' error : '+Msg);
 end;

Function TNWHostPrinter.GetError:integer;
 begin
  Result:=fError;
  fError:=0;
 end;

Function IpToLong(ip:string):LongInt;
 var
  x,i:byte;
  ipx:array[0..3] of byte;
  v:integer;
 begin
  Result:=0;
  longint(ipx):=0; i:=0;
  for x:=1 to length(ip) do
   if ip[x]='.' then begin
    inc(i);
    if i=4 then exit;
   end else begin
    if not (ip[x] in ['0'..'9']) then exit;
    v:=ipx[i]*10+ord(ip[x])-ord('0');
    if v>255 then exit;
    ipx[i]:=v;
   end;
  result:=longint(ipx);
 end;

Function HostToLong(AHost:string):LongInt;
 Var
  Host:PHostEnt;
 begin
  Result:=IpToLong(AHost);
  if Result=0 then begin
   Host:=GetHostByName(PChar(AHost));
   if Host<>nil then Result:=integer(pointer(host^.h_addr_list^)^);
  end;
 end;

Procedure TNWHostPrinter.Connect;
 var
  SockAddr:TSockAddr;
 begin
  if Active then exit;
  if fState>0 then SetError(1,'Invalid status');

  if fHostName='' then SetError(2,'Missing HostName');

  fHandle:=AllocateHwnd(WndProc);

  fSocket:=socket(AF_INET,SOCK_STREAM,0);
  if fSocket=INVALID_SOCKET then SetError(3,'Winsock error');

  wsaAsyncSelect(fSocket,fHandle,wm_user,(FD_CONNECT or FD_READ or FD_CLOSE));

  // Call Host
  FillChar(SockAddr,SizeOf(SockAddr),0);
  SockAddr.sa_family:=AF_INET;
  SockAddr.sin_port :=swap(23);
  SockAddr.sin_addr.s_addr:=HostToLong(fHostName);
  if (winsock.Connect(fSocket,SockAddr,SizeOf(SockAddr))<0)
  and(wsaGetLastError=10035) then begin
   fState:=1;
   fActive:=True;
  end else begin
   CloseSocket(fSocket);
   SetError(4,'Connect error');
  end;
 end;

Procedure TNWHostPrinter.Disconnect;
 begin
  if not Active then exit;
  ShutDown(fSocket,1);  // stop sending , wait a "Disconnect" message
 end;

Procedure TNWHostPrinter.Disconnected;
 var
  dummy:array[0..1024] of char;
 begin
  repeat until recv(fSocket,dummy,1024,0)<=0; // read pending data
  CloseSocket(fSocket);
  fSocket:=INVALID_SOCKET;

  DeallocateHWnd(fHandle);

  fState:=0;
  fActive:=False;

  if Assigned(EOnClose) then try EOnClose(Self) except on e:exception do showmessage(e.message) end;
 end;

Function TNWHostPrinter.Setup:boolean;
 begin
  with TNWHostPrinterSetupForm.Create(Application) do begin
   GetPrinterInfo(Self);
   Result:=(ShowModal=mrOk);
   if Result then SetPrinterInfo(Self);
   Free;
  end;
 end;

Procedure TNWHostPrinter.WndProc(Var Msg:TMessage);
 begin
  with msg do
   if msg=wm_user then
    if LParamHi=0 then begin
     case lParamLo of
      $0001 : ReadData(wParam);
      $0010 : if Assigned(EOnConnect) then EOnConnect(Self);
      $0020 : Disconnected;
     end;
    end else begin
//     Disconnect;
     SetError(lParamHi,'WndProc socket error');
    end
   else
    Result:=DefWindowProc(fHandle, Msg, wParam, lParam);
 end;

Procedure TNWHostPrinter.ReadData(Socket:integer);
 var
  l:integer;     // length of incomming data
  s:ShortString; // incomming data string
  i:integer;     // parsing index
  c:char;        // current parsing char
  r:string;      // reply string

 function StringIndex(s:string;a:array of pchar):integer; // return index of "s" in "a" (or -1)
  begin
   Result:=High(a);
   while Result>=0 do begin
    if s=a[Result] then exit;
    dec(Result);
   end;
  end;

 function PaperSource(Name:string):string; // get papersource code
  begin
   case StringIndex(Name,Papers) of
    0:Result:=#$FF#$FF; // double escape
    1:Result:=#$00;
    2:Result:=#$02#$01;
    3:Result:=#$02#$02;
    4:Result:=#$02#$03;
    5:Result:=#$04;
    6:Result:=#$05;
    7:Result:=#$06;
    8:Result:=#$07;
    9:Result:=#$08;
   10:Result:=#$0E;
   11:Result:=#$0F;
   12:Result:=#$10;
   else Result:=#$FF#$FF;
   end;
  end;

 function EnvSource(Name:string):string;
  begin
   case StringIndex(Name,Envelops) of
    0:Result:=#$FF#$FF;
    1:Result:=#$00;
    2:Result:=#$06;
    3:Result:=#$09;
    4:Result:=#$0A;
    5:Result:=#$0B;
    6:Result:=#$0C;
    7:Result:=#$0D;
   else Result:=#$FF#$FF;
   end;
  end;

 begin
  l:=recv(Socket,s[1],255,0);
  r:='';
  while l>0 do begin
   s[0]:=chr(l);
   for i:=1 to length(s) do begin
    c:=s[i];
    case fState of
    // data or escape char
     1 : if c=tnIAC then fState:=2 else fData:=fData+c;
    // escape sequence
     2 : case c of
         // double escape to add escape char to data flow
          tnIAC : begin fData:=fData+c; fState:=1 end;
          tnDO  : fState:=3;
          tnWILL: fState:=4;
          tnSB  : begin fNegoce:=''; fState:=5; end;
         // end of data, process data
          tnEND : begin
                  fState:=1;
                  r:=r+Execute;
                  fData:='';
                 end;
// tnSE  : Status:='Unexpected SE ?!'
         // unknow escape sequence
          else begin SetError(90,'<!> IAC #'+IntToStr(ord(c))); {Disconnect;} exit; end;
         end;
    // DO
     3 : begin
          fState:=1;
          if c in [tnBIN,tnTERMTYPE,tnEOR,tnNEWENV] then
           r:=r+tnIAC+tnWILL+c
          else
           r:=r+tnIAC+tnWONT+c;
         end;
    // WILL
     4 : begin
          fState:=1;
          if c in [tnBIN,tnECHO,tnSUPPRESS_GA,tnEOR,tnNEWENV] then
           r:=r+tnIAC+tnDO+c
          else
           r:=r+tnIAC+tnDONT+c;
         end;
    // SubNegoce
     5 : begin
          fNegoce:=fNegoce+c;
         // end of SubNegoce
          if c=tnSE then begin
           fState:=1;
          // Terminal type
           if fNegoce=tnTERMTYPE+ttSEND+tnIAC+tnSE then begin
            r:=r+tnIAC+tnSB+tnTERMTYPE+ttIS+ TERM +tnIAC+tnSE;
           end else
          // New Environnemnt
           if fNegoce[1]=tnNEWENV then begin
            r:=r+tnIAC+tnSB+tnNEWENV+ttIS;
            r:=r+tnVAR+'DEVNAME'+tnVAL+fDevName;
            r:=r+tnVAR+'IBMMSGQNAME'+tnVAL+GetName(fMSGQ);
            r:=r+tnVAR+'IBMMSGQLIB' +tnVAL+GetLib(fMSGQ);
            r:=r+tnVAR+'IBMTRANSFORM'+tnVAL+'1';
            r:=r+tnVAR+'IBMMFRTYPMDL'+tnVAL+fModel;
            if fModel='*WSCST' then begin
             r:=r+tnVAR+'IBMWSCSTNAME'+tnVAL+GetName(fWSCST);
             r:=r+tnVAR+'IBMWSCSTLIB' +tnVAL+GetLib(fWSCST);
            end;
            r:=r+tnVAR+'IBMFONT'     +tnVAL+fFont;
            r:=r+tnVAR+'IBMFORMFEED' +tnVAL+fFormFeed;
            r:=r+tnVAR+'IBMPPRSRC1'  +tnVAL+PaperSource(fPaper1);
            r:=r+tnVAR+'IBMPPRSRC2'  +tnVAL+PaperSource(fPaper2);
            r:=r+tnVAR+'IBMENVELOPE' +tnVAL+EnvSource(fEnvelope);
            r:=r+tnIAC+tnSE;
           end else begin
            SetError(10,'<!> SubNegoce #'+IntToStr(ord(fNegoce[1]))+ ' '+fNegoce);
           fNegoce:='';
          end;
         end;
      end;
     end;
   end;
   if r<>'' then begin
    send(Socket,r[1],length(r),0);
    r:='';
   end;
   l:=recv(Socket,s[1],255,0);
  end;
 end;

Function TNWHostPrinter.Execute:string;
 var
  Head:packed record
   Size:word; // head + data
   Kind:word; // $A012 - WARNING ! can change in future version
   Flow:word; // Data Flow
  end;
 begin
  Result:='';
  if Length(fData)<SizeOf(Head) then SetError(90,'Invalide Data Header');
  Move(fData[1],Head,SizeOf(Head));
  Delete(fData,1,SizeOf(Head));
  if Head.Kind<>$A012 then SetError(99,'Invalide Header found !');
  Case Head.Flow of
   $0090 : StartupInfo;
   $0101 : Result:=PrintRecord;
   else    SetError(90,'Unknow data flow #'+IntToStr(Head.flow));
  end;
 end;

Procedure TNWHostPrinter.StartupInfo;
 var
  x:integer;
  s:string;
 begin
  x:=ord(fData[1])+2;
  s:=EbcdicToAnsiW(Copy(fData,x+4,4));
  fRemoteName:=Trim(EbcdicToAnsiW(Copy(fData,x+16,10)))+'/'+Trim(EbcdicToAnsiW(Copy(fData,x+8,8)));
  if s[1]='I' then s[1]:='1';
  case StrToInt(s) of
   1901 : Status:='Virtual device has less function than source device.';
   1902 : Status:='Dmarre.';
   1906 : Status:='Automatic sign-on requested, but not allowed.';
   2702 : Status:='Device description not found.';
   2703 : Status:='Controller description not found.';
   2777 : Status:='Damaged device description.';
   8901 : Status:='Device not varied on.';
   8902 : Status:='Device not available.';
   8903 : Status:='Device not valid for session';
   8906 : Status:='Session initiation failed.';
   8907 : Status:='Session failure.';
   8910 : Status:='Controller not valid for session.';
   8916 : Status:='No matching device found.';
   8917 : Status:='Not authorized to object.';
   8918 : Status:='Job canceled.';
   8920 : Status:='Object partially damaged.';
   8921 : Status:='Communications error.';
   8922 : Status:='Negative response received.';
   8923 : Status:='Start-up record built incorrectly.';
   8925 : Status:='Creation of device failed.';
   8928 : Status:='Change of device failed.';
   8929 : Status:='Vary on or vary off failed.';
   8930 : Status:='Message queue does not exist';
   8934 : Status:='Start-up for S/36 WSF received.';
   8935 : Status:='Session rejected.';
   8936 : Status:='Security failure on session attempt.';
   8937 : Status:='Automatic sign-on rejected.';
   8940 : Status:='Automatic configuration failed or not allowed.';
   1904 : Status:='Source system at incompatible release.';
   else   Status:='Unknow error #'+s;
  end;
  hPrinter:=0;
  ibmcode:=0;
 end;

Procedure TNWHostPrinter.SetStatus(Value:string);
 begin
  fStatus:=Value;
  if Assigned(EOnChange) then EOnChange(Self);
 end;


Function TNWHostPrinter.GetPrinting:boolean;
 begin
  Result:=(hPrinter<>0);
 end;

Const
 PrintComplet=#$00#$0A#$12#$A0#$01#$02#$04#$00#$00#$01#$FF#$EF;
 PrintError  =#$00#$0A#$12#$A0#$01#$02#$04#$81#$81#$01#$FF#$EF;
{
   This would normally follow receipt of a print record, such as the
   example in Figure 4.  This indicates successful completion of a print
   request.

   +--------------------------------------------------------------------+
   |   +-- Length of structure (LLLL)                                   |
   |   |    +-- GDS identifier                                          |
   |   |    |    +-- Data flow record                                   |
   |   |    |    |   +-- Length of pass-through specific header (LL)    |
   |   |    |    |   |   +-- Flags                                      |
   |   |    |    |   |   |   +-- Printer operation code                 |
   |   |    |    |   |   |   |                                          |
   | +--+ +--+ +--+ ++ +--+ ++                                          |
   | |  | |  | |  | || |  | ||                                          |
   | 000A 12A0 0102 04 0000 01                                          |
   +--------------------------------------------------------------------+
    Figure 5. Client sending server a print complete record

   - '000A'X = Logical record length, including this byte (LLLL)
   - '12A0'X = GDS LU6.2 header
   - '0102'X = Data flow response record (client to server)
   - '04'X   = Length of pass-through specific header (LL)
   - '0000'X = Good Response
   - '01'X   = Print Complete
 }

Function WindowsError:string;
 var
  i:integer;
 begin
  SetLength(Result,1024);
  i:=FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil, GetLAstError,LOCALE_USER_DEFAULT,PChar(Result),1024,nil);
  SetLength(Result,i);
 end;

function TNWHostPrinter.OpenPrinter:boolean;
 var
  i,r:integer;
  s,ss:string;
  Host:string;
 begin
  Result:=False;
  try
 // Context
  ss:='';
  s:='NWDSCreateContextHandle';
  if NWDSCreateContextHandle(fContext)<>0 then exit;
  s:='NWDSGetContext';
  if NWDSGetContext(fContext, DCK_FLAGS,i)<>0 then exit;
  i:=i or DCV_TYPELESS_NAMES;
  s:='NWDSSetContext';
  if NWDSSetContext(fContext, DCK_FLAGS,i)<>0 then exit;
 // Connexion
  s:='NWReadString - Host Server "'+fPrinter+'"';
  ss:=' printer="'+fPrinter+'"';
  if NWReadString(fContext,fPrinter,'Host Server',Host)<>0 then exit;
  ss:=ss+' host="'+Host+'"';
  if Host<>'' then begin
   s:='NWReadString - CN "'+Host+'"';
   ss:=ss+' CN="'+Host+'"';
   if NWReadString(fContext,Host,'CN',Host)<>0 then exit;
  end;
  if Host<>'' then begin
   s:='NWCCOpenConnByName';
   if NWCCOpenConnByName(0,PChar(Host),NWCC_NAME_FORMAT_BIND, NWCC_OPEN_LICENSED,NWCC_TRAN_TYPE_WILD, fConnect)<>0 then exit;
  end else begin
   s:='NWCCGetPrimConnRef';
   if NWCCGetPrimConnRef(r)<>0 then exit;
   s:='NWCCOpenConnByRef';
   if NWCCOpenConnByRef(r,NWCC_OPEN_LICENSED, NWCC_RESERVED, fConnect)<>0 then exit;
  end;
 // Queue ID
  s:='NWDSMapNameToID';
  if NWDSMapNameToID(fContext, fConnect, Pchar(fPrinter),fQueue)<>0 then exit;
 // Job
  FillChar(fJob,SizeOf(fJob),0);
  with fJob do begin
  // set required fields ...
   targetServerID:=-1;  // any print server
   FillChar(targetExecutionTime, sizeOf(targetExecutionTime), $FF);    // ASAP
   jobControlFlags := QF_ENTRY_OPEN + QF_ENTRY_RESTART;// + QF_USER_HOLD;
  // set optional fields ...
   StrPLCopy(jobDescription,DevName,SizeOf(jobDescription));
  // Client Record area
//   version :=1; 0 ?
   tabSize :=8;
   numberOfCopies:=1;
   printControlFlags:=NO_FORM_FEED;
//   maxLinesPerPage:=66;
//   maxCharsPerLine:=132;
//   formName:='NWHPrint';
//   bannerNameField:='BannerName';
//   bannerFileField:='BannerFile';
//   headerFileName :='HeaderName';
  end;
 // Open Queue File
  s:='NWCreateQueueFile2';
  if NWCreateQueueFile2(fConnect,fQueue,fJob,hPrinter)<>0 then exit;
  Result:=True;
  finally
   if not result then fErrorMsg:='Erreur Netware ('+s+ss+')';
  end;
 end;

function TNWHostPrinter.WritePrinter(s:string):boolean;
 begin
  Result:=FileWrite(hPrinter,s[1],Length(s))=Length(s);
 end;

function TNWHostPrinter.AbortDoc:boolean;
 begin
  if hPrinter=0 then
   Result:=True
  else begin
   Result:=NWCloseFileAndAbortQueueJob2(fConnect,fQueue,fJob.jobNumber,hPrinter)=0;
   NWCCCloseConn(fConnect);
   NWDSFreeContext(fContext);
   hPrinter:=0;
  end;
 end;

function TNWHostPrinter.EndDoc:boolean;
 begin
  Result:=NWCloseFileAndStartQueueJob2(fConnect,fQueue,fJob.jobNumber,hPrinter)=0;
  NWCCCloseConn(fConnect);
  NWDSFreeContext(fContext);
  hPrinter:=0;
 end;

Function TNWHostPrinter.PrintRecord:string;
 var
  x:integer;
 begin
  Result:=PrintComplet;
  x:=ord(fData[1])+1; // data offset
  while x<=length(fData) do begin // until end of buffer
   if (ibmcode=0)and(fData[x]=#0) then begin // found #0 outside an escape sequence
    inc(x);
    EndDoc;
    Status:='READY.';
    if Assigned(EOnPrint) then EOnPrint(Self);
   end else begin
    if ibmcode=0 then begin
     if fData[x]<>#3 then ShowMessage('0x03 not found :('); inc(x);
     ibmcode:=ord(fData[x]); inc(x);
     hpcl:='';
    end;
    while (x<=length(fData))and(ibmcode>0) do begin
     hpcl:=hpcl+fData[x]; inc(x);
     dec(ibmcode);
     if ibmcode=0 then begin
      if hPrinter=0 then begin
       OpenPrinter;
       if hPrinter=0 then begin
        Result:=PrintError;
        //fErrorMsg:='Impossible d''ouvrir l''imprimante rseau';//WindowsError;
        if Assigned(EOnPrintError) then EOnPrintError(self);
        exit;
       end;
       Status:='Printing...';
       if Assigned(EOnPrint) then EOnPrint(Self);
      end;
      if not WritePrinter(hpcl) then begin
       AbortDoc;
       Result:=PrintError;
       fErrorMsg:='Erreur d''criture sur l''imprimante';//WindowsError;
       if Assigned(EOnPrintError) then EOnPrintError(self);
       exit;
      end;
     end;
    end;
   end;
  end;
 end;

procedure WSAStartup;
 var
  WSA:TWSAData;
 begin
  WINSOCK.WSAStartup($101,WSA);
 end;

Initialization WSAStartup;
Finalization   WSACleanup;

end.
